import { MessageTransmissionTransmitter } from "./msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "./msg.transmission.receiver";
import { ConnectionManager } from "../connector/connector.manager";
import { EventMessage, MessageTransmission, MessageTransmissionManager as MessageTransmissionManagerInterface, ReceiverProfile, TransmissionProfile, TransmitterProfile } from "../interface/transport.interface";
import { v4 as uuidv4 } from 'uuid'
import { AdapterSet, AdaptorTransmissionRole, RequestResponseConnectionAdapter, Transport, TransportEvent, TransportService } from "../interface/connector.interface";
import { MessageTransmissionRequestResponse } from "./msg.transmission.request-response";
import { filter, Observable, Observer, Subject } from "rxjs";
import { WebsocketTransportService } from "../transport/websocket";
import { HttpTransportService } from "../transport/http";
import { error } from "console";

export class MessageTransmissionManager implements MessageTransmissionManagerInterface{
    private transportServiceArray: TransportService[] = []
    private transportSet: Set<TransportSet> = new Set()
    transmission: MessageTransmission[] = []
    connectionManager!: ConnectionManager
    event!: Subject<TransportEvent>

    constructor(event: Subject<TransportEvent>) {
        // logic here
        console.log(`TransmissionManager: Contructing Transmission Manager...`)
        this.event = event
        this.connectionManager = new ConnectionManager(this.event)

        this.sort(this.transportSet)
        this.transportSet.forEach(set => {
            this.setUpTransportService(set, event)
        })
    }

    /* so there will be some changes here. will nto be assigning just one, but all of them dynamically to pour into this boy
    Transmitter only have to call this once. */
    subscribe(): Observable<MessageTransmission> {
        return new Observable((observer: Observer<MessageTransmission>) => {
            this.event.pipe(
                filter(event => event.event == 'New Client')
            ).subscribe(event => {
                // get all adapters for all the connection
                let adapterSet: AdapterSet[] = []
                let clientId: string = (event.data as EventMessage).clientId
                if (this.transportServiceArray.length > 0) {
                    this.transportServiceArray.forEach(transport => {
                        adapterSet.push(this.connectionManager.getAdapter(clientId, transport))
                    })
                } else {
                    observer.error('No Transport is instantiated.... ERROR!!!!')
                }

                // 1 set only
                let transmitter: MessageTransmissionTransmitter = this.getTransmitter(clientId, adapterSet, this.event.asObservable())
                let receiver: MessageTransmissionReceiver = this.getReceiver(clientId, adapterSet, this.event.asObservable())
                let requestResponse: MessageTransmissionRequestResponse = this.getRequestResponse(transmitter, receiver, this.event.asObservable())
                let transmission: MessageTransmission = {
                    id: clientId,
                    transmitter: transmitter,
                    receiver: receiver,
                    requestResponse: requestResponse,
                    event: this.event.asObservable()
                }
                this.transmission.push(transmission)

                observer.next(transmission)
            })
        })
    }


    private getTransmitter(transmissionId: string, adapterSets: AdapterSet[], event: Observable<TransportEvent>): MessageTransmissionTransmitter {
        let transmitterProfile: TransmitterProfile = {
            id: transmissionId,
            name: '', // for now make it empty. We will use the assigned uuid here
            dateCreated: new Date()
        }
        return new MessageTransmissionTransmitter(transmitterProfile, adapterSets, event)
    }

    private getReceiver(transmissionId: string, adapterSets: AdapterSet[], event: Observable<TransportEvent>): MessageTransmissionReceiver {
        let receiverProfile: ReceiverProfile = {
            id: transmissionId,
            name: '', // for now make it empty. We will use the assigned uuid here
            dateCreated: new Date()
        }
        return new MessageTransmissionReceiver(receiverProfile, adapterSets, event)
    }

    private getRequestResponse(transmitterInstance: MessageTransmissionTransmitter, receiverInstance: MessageTransmissionReceiver, event: Observable<TransportEvent>): MessageTransmissionRequestResponse {
        return new MessageTransmissionRequestResponse(transmitterInstance, receiverInstance, event)
    }


    // Server to be set up as well as acquiring client information if needed. Like in the case for grpc and socket. Http not requ`ired.
    private setUpTransportService(transportSet: TransportSet, event: Subject<TransportEvent>): void {
        this.instantiateTransportService(transportSet.transport, event).then((transportService: TransportService) => {
            this.transportServiceArray.push(transportService)
            if (transportService instanceof WebsocketTransportService) {
                console.log(`Just Double Checking... this is websocket`)
                transportService.startServer(transportSet.port);
            } else if (transportService instanceof HttpTransportService) {
                console.log(`Just Double Checking... this is http`)
                transportService.startServer(transportSet.port);
                // Additional Http-specific setup if needed.
            }
        }).catch((error) => { throw new Error(error) })
    }

    private async instantiateTransportService(transportType: Transport, event: Subject<TransportEvent>): Promise<TransportService> {
        return new Promise((resolve, reject) => {
            if (transportType == Transport.Websocket) {
                resolve(new WebsocketTransportService(event))
            }
            else if (transportType == Transport.Http) {
                resolve(new HttpTransportService(event))
            } else {
                reject(`No Transport Service is not properly instantiated`)
            }
        })
    }

    private sort(transportSet: Set<TransportSet>): void {
        let transportList: string[] = process.env.Transport?.split(',') || []
        let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
        transportList.forEach((transport, index) => {
            transportSet.add({ transport: transport, port: portList[index] } as unknown as TransportSet)
        })
    }
}


interface TransportSet {
    transport: Transport,
    port: number
}

